;JT65_Beacon_LMX2541.ASM    G4JNT  DEc 2012
;JT65 Symbol generator for driving LMX24xxl FractN synth with direct register programming
; Sets up all synth registers
;Derived from LMX24xx version.   32 bit output words
; Fosc = 10MHz (2.5MHz clock)   Can be changed
;  CW Keying Line reinstated
;30s CW message added
;Reprogrammes frequency grid from EE for accurate carrier and JT frequencies
;EE storage problems - some fixed Synth setup data set as constants


        LIST P = 16F628

        INCLUDE "P16F628.INC"
    __config    0x3F61      ;LV Programming off , OSC XT
    __idlocs    0x8027      ;


 	#define		SynthData	PORTA, 0
	#define		SynthClk	PORTA, 1
	#define		SynthLE		PORTA, 2
	#define     	KEYLINE     	PORTA, 3
    
	#define     	PPS         	PORTB, 0
	#define     	GPSIN       	PORTB, 1
	#define     	GPSOUT      	PORTB, 2
	#define     	TESTLINK    	PORTB, 3
	#define     	LEDGrn     	PORTB, 4
	#define     	LEDRed      	PORTB, 5


	#define     	TestModeFlag	Flags, 1
	#define     	GPSOk           Flags, 2
	#define     	CurrData        Flags, 4
	#define     	HexFlag         Flags, 6
	#define		AllowData	Flags, 0
	#define     	AllowSendFlag	Flags, 7


	SERIALBITS	=	d'32'	;32 for LMX2541 synth
	GPSDataType	= 	1	;1 = NMEA  0 = Motorola Binary
	NMEAPol		=	1	;1 = RS232 Levels, 0 = TTL
	BOTHMINUTES	=	1	;1 = every minute, 0 = Alternate (even) minutes
	IGNOREPPS	=	0	;1 = Serial data timing only (less accurate), 0 = use PPS signal as well

	CWSPEED		=	d'80'   ;CW Dot length, ms
	BAUD9600	=	d'40'	;6.N + 18 = Fc / Baud   or   N ~ (Fxtal / Baud)/24 - 3
	BAUD4800	=	d'83'		
	JT65CONST	=	d'907'	 ;Interrupt frequency / 2.691

	GRIDNUMBER	=	3	;Number of registers needed to set frequency grid

  errorlevel 1, -207  ;Suppress msgs. and warnings about labels after col 1.
;=======================================================================
  org   0x2100

ConfigData		;32 bit words sent in sequence  for registers that remain unchanged
 ;  Fout 1296.78MHz  Resolution 10.Hz
  ;  R =  1      D =  500000      N =  259      F =  178000
;   de  0x00, 0x00, 0x00, 0x17    ;  Reg7   	;These sent as constants in SendfromEE routine
;   de  0x00, 0x00, 0x00, 0xCD    ;  Reg13   	;  Check Reg13 VCO_DIV_OPT setting
;   de  0x00, 0x00, 0x00, 0x1C    ;  Reg12   
;   de  0x28, 0x00, 0x14, 0x09    ;  Reg9   
    de  0x01, 0x11, 0xCE, 0x58    ;  Reg8   
   de  0x00, 0x1F, 0x33, 0x26    ;  Reg6   
   de  0xA0, 0x04, 0x00, 0x05    ;  Reg5   
   de  0xFF, 0x48, 0x20, 0xA4    ;  Reg4   
   de  0x00, 0xA8, 0x73, 0x03    ;  Reg3   
;   de  0x04, 0x38, 0xB0, 0x82    ;  Reg2   
;   de  0x00, 0xC0, 0x00, 0x11    ;  Reg1   
;   de  0x58, 0x3D, 0x15, 0x90    ;  Reg0   
EndConfigData


JTFreqData				;10.7666Hz
   de  0x04, 0x71, 0x60, 0xF2    ;  Reg2   
   de  0x00, 0x80, 0x00, 0x11    ;  Reg1   
   de  0x85, 0xFA, 0x10, 0x30    ;  Reg0   

EndJTFreqData

CarrierFreqData				;10Hz
   de  0x04, 0x7A, 0x12, 0x02    ;  Reg2   
   de  0x00, 0x80, 0x00, 0x11    ;  Reg1   
   de  0xB7, 0x50, 0x10, 0x30    ;  Reg0   
EndCarrierFreqData

JT65MsgData
  include	"JT65SYMB.INC"     ;JT65 tones

CWMsg	
	de      " ",0
;=================================================================================						   
  org     0
    nop
    movlw   b'00000000'     ;No interupts until passed thro here once
    movwf   INTCON          ;
    goto    StartUp         ;jump to main code

;------------------ Interrupt routine, get here 4 cycles after TMR overflow --
ints                            ;/4 prescale &  /256 for ovflow int  = Interrupt Rate
	bcf     INTCON , T0IF   ;
	movwf   Wreg            ;Save W
	swapf   STATUS , W      ; nb.  movwf does not affect any STATUS bits
	movwf   SaveSts         ;Stave STATUS, without affecting it

	incf    IntCountLo
	btfsc   STATUS, Z
	incf    IntCountHi

        movlw   HIGH(JT65CONST)    ;Int rate/JT65CONST = 2.691337Hz symbol rate
	subwf   IntCountHi, W
	btfss   STATUS, Z
	goto    OutInterrupt

	movlw   LOW(JT65CONST)
	subwf   IntCountLo, W
	btfss   STATUS, Z
	goto    OutInterrupt
	   ;............             Here at 2.69Hz
	clrf    IntCountLo
	clrf    IntCountHi

	btfsc   SymbCount, 0    ;Flash Red LED at half symbol rate
	bsf     LEDRed
	btfss   SymbCount, W
	bcf     LEDRed

	movf    SymbCount       ;Test for First symbol, in which case
	btfsc   STATUS, Z       ;  don't do any DDS update
	goto    NotFirstPass

NotFirstPass

	movlw   d'126'
	subwf   SymbCount, W    ;So do the testing for a full message here, 
	btfss   STATUS, Z
	goto    NotLastSymbol

	bcf	AllowData	
	bcf     INTCON, T0IE    ;Inhibit interrupt after final symbol
	goto    OutInterrupt    ;  and don't do anything else
	;...............
NotLastSymbol
		movf    SymbCount       ;Test for first symbol
		btfsc   STATUS, Z
		call    SyncToStack     ;Load sync pattern into stack first time here

		rlf     Stack0          ;Stack holds sync / message tone pattern
		rlf     Stack1    
		rlf     Stack2
		rlf     Stack3          
		rlf     Stack4
		rlf     Stack5
		rlf     Stack6
		rlf     Stack7
		rlf     Stack8
		rlf     Stack9
		rlf     StackA
		rlf     StackB
		rlf     StackC
		rlf     StackD
		rlf     StackE
		rlf     StackF
		btfss   STATUS, C       ;1 = sync, 0 = character
		goto    JTSymb

		clrf	FreqCache            ;0 passed to routine sends sync tone
		bsf	AllowSendFlag
		goto    SendTone
	;...................
JTSymb                          ;Get next symbol value in message (0-63)
	movf    CharCount, W
	addlw   LOW(JT65MsgData)
	call    GetEE           ;Get value 0 to 63
        addlw   2               ;2* interval Offset from sync (was 3 in JT44)
	movwf	FreqCache
	bsf	AllowSendFlag   ;Pass symbol value to routine,
								; generates frequency data in D0-D3
	incf    CharCount
	;..........
SendTone                        ;Get here with frequency data in D0-5
	incf    SymbCount       ;Increment here, but tested at start allows 
								; the DDS update at last+1 interval, ie. =136
OutInterrupt
	swapf   SaveSts , W     ;Restore STATUS
	movwf   STATUS
	swapf   Wreg            ;Restore W
	swapf   Wreg , W        ;Can't use movf as it  affects Z bit

	retfie    

;=====================================
StartUp 
    clrf    PORTA       ;Special setup to disable comparator on 'F628
    movlw   7
    movwf   CMCON
	
    bsf     STATUS,RP0      ;ram page 1
    movlw   b'00000001'     ;Internal input, Prescalar /4  pull-ups
    movwf   OPTION_REG
    movlw   b'00000000'     ;A0/2 Outputs, A3 Out to GPS
    movwf   PORTA           ;
    movlw   b'00001011'     ;B0 PPS, B1 GPSIN, B2 GPSOUT B3 TestLink  B4/5 LEDs ,
    movwf   PORTB           
    bcf     STATUS,RP0      ;ram page 0
    clrf    PORTB
    clrf    PORTA	 

    call    LongDelay		;Allow GPS Supppy to stabilise
    bsf     LEDGrn
    call    SetUpGPS	

    call    SendFromEE	    ;Initialise other registers
    movlw   LOW(CarrierFreqData)
    call    SetGrid

    clrf    TMR0
    clrf    IntCountHi
    clrf    IntCountLo
    clrf    Flags
    clrf    CharCount
    clrf    SymbCount

    bsf	    KEYLINE
    bsf     LEDRed
    bcf     LEDGrn
    call    LongDelay
    bcf     LEDRed
    bsf     LEDGrn
    call    LongDelay
    bsf     LEDRed
    bcf     LEDGrn
    call    LongDelay

    movlw   b'10000000'     ;Enable global int, not timer overflow int yet
    movwf   INTCON

WaitGPSFirstLock            ;
                           
    call    GetGPSData      
    btfss   GPSOk
    goto    WaitGPSFirstLock

    bcf     LEDRed
;...........................
MainLoop
    call    GetGPSData      ;Sets Minutes, Seconds and RxStatus

  if BOTHMINUTES == 0
    btfsc   Minutes, 0      ;Test for the odd minute
    goto    CarryOn
  endif

    movf    Seconds, W
    sublw   d'00'
    btfss   STATUS, Z
    goto    CarryOn 
                      ;Now at even minute : 00


  if IGNOREPPS == 1
	call	LongDelay
  else
	call    WaitPPS     ;Start on the even minute boundary + 1s
  endif

	clrf    CharCount    
	clrf    SymbCount    ;Re-Initialise registers
	clrf    IntCountLo
	clrf    IntCountHi

	movlw	d'0'
	movwf	LastFreq		;Changing the grid sets Tone 0 startup

	bcf	KEYLINE			;Stop wideband transients
	call	CWDelay
	movlw 	LOW(JTFreqData)		;change to JT65 Frequency grid
	call	SetGrid
	call	CWDelay
	call	CWDelay
	bsf	KEYLINE

	call	LongDelay

	bsf     INTCON, T0IE    	;Enable the interrupt, starts JT65 sequence

JT65Loop
	btfss   AllowSendFlag
	goto    NoSend
	movf    FreqCache, W
	call    SendFreq		;Has to be done outside interrupt routine
	bcf	AllowSendFlag		; as an interrupt overrun would occur
NoSend
	btfsc   INTCON, T0IE ;Sit in here waiting for completion   
	goto    JT65Loop

	movlw   0
	call    SendFreq

	call    LongDelay
	call	LongDelay		;Give a few seconds of sync tone to clear as the decoder seems to get upset

	bcf     KEYLINE			;Carrier off before changing frequency grid settings
	call	Delay1ms
	call	Delay1ms
	movlw   LOW(CarrierFreqData)
	call    SetGrid

	movlw   LOW(CWMsg)		;carrier turned back on in CW message after Delay
	movwf   CWMsgCnt
	call    SendCWMsg
CarryOn
	movlw	d'30'
	subwf	Seconds, W
	btfss	STATUS, Z	
	goto	MainLoop
	movlw	LOW(CWMsg)
	movwf	CWMsgCnt
	call	SendCWMsg
	goto    MainLoop
;.............................
NoTxValue  
    movlw   4
    movwf   Counter     ;Unique flash pattern to show Random value
NoTxLoop                ;  when transmit sequence has NOT been triggered
    bcf     LEDGrn      ; Flashes R+G 
    bsf     LEDRed
    call    CWDelay
    bsf     LEDGrn
    bcf     LEDRed
    call    CWDelay
    bcf     LEDGrn
    call    CWDelay
    call    CWDelay
    call    CWDelay
    call    CWDelay
    call    CWDelay

    decfsz  Counter
    goto    NoTxLoop

    goto    MainLoop
;===========================================

;--------------------------------

WaitPPS                 ;In case PPS is still high
    btfsc   PPS         ; look for a low to high transition
    goto    WaitPPS

WaitRefPPS              ;Now wait for the rising edge itself
    btfss   PPS        
    goto    WaitRefPPS 

    return
;----------------------
Multiply32  ;32 bit * 32 bit multiplication, more efficient method
    Clrf    Q0        ;Input A0/1/2/3 and B0/1/2/3, output Q0/2/3/4/5/6/7/8
    clrf    Q1        ;A destroyed,  B kept
    clrf    Q2        
    clrf    Q3       
    clrf    Q4
    clrf    Q5
    clrf    Q6
    clrf    Q7

    movlw   d'32'       ;A and B 24 bit
    movwf   Counter
MultLoop            
    bcf     STATUS, C
    rrf     A3
    rrf     A2          ;Working bit into carry
    rrf     A1
    rrf     A0          ;Effectivley adds shifted version of B
    btfss   STATUS, C   ; into Acc depending on working bit of A
    goto    NoMult      ; by shifting the accumulator instead of B itself

    movf    B0, W       ;get shifted B0 value
    addwf   Q4        ; and add into accumulator if 1 in 
    btfss   STATUS, C   ; appropriate bit in Avalue
    goto    MulAdd1
    incf    Q5
    btfss   STATUS, Z
    goto    MulAdd1     
    incf    Q6
    btfss   STATUS, Z
    goto    MulAdd1
    movlw   1           ;Need more complex addition routine for last Acc as
    addwf   Q7        ; final Carry has to be preserved for the shift
MulAdd1
    movf    B1, W
    addwf   Q5
    btfss   STATUS, C
    goto    MulAdd2
    incf    Q6
    btfss   STATUS, Z
    goto    MulAdd2
    movlw   1
    addwf   Q7

MulAdd2
    movf    B2, W
    addwf   Q6
    btfss   STATUS, C
    goto    MulAdd3
    movlw   1
    addwf   Q7
MulAdd3
    movf    B3, W
    addwf   Q7

NoMult                  ;Need to rotate in carry from final addition.  C is
    rrf     Q7        ; already zero if we've jumped in here.
    rrf     Q6
    rrf     Q5        
    rrf     Q4
    rrf     Q3
    rrf     Q2
    rrf     Q1
    rrf     Q0
    decfsz  Counter
    goto    MultLoop

    return              ;Q7/7/6/5/4/3/2/1 = A * B     
;------------------------------------------------------------------
SyncToStack
	movlw   0x98
	movwf   StackF
	movlw   0xFD
	movwf   StackE
	movlw   0x45
	movwf   StackD
	movlw   0x91
	movwf   StackC
	movlw   0xCF
	movwf   StackB
	movlw   0x6F
	movwf   StackA
	movlw   0x1A
	movwf   Stack9
	movlw   0xB3
	movwf   Stack8
	movlw   0x54
	movwf   Stack7
	movlw   0x81
	movwf   Stack6
	movlw   0x80
	movwf   Stack5
	movlw   0xD2
	movwf   Stack4
	movlw   0xD5
	movwf   Stack3
	movlw   0x32
	movwf   Stack2
	movlw   0x43
	movwf   Stack1
	movlw   0xFC
	movwf   Stack0

	return
;---------------------------------------
GetEE         ;different from ;F84   EE registers all in bank 1 now 
	bsf     STATUS,RP0      ;ram page 1
	movwf   EEADR
	bsf     EECON1 , RD
	movf    EEDATA , W
	bcf     STATUS , RP0    ;ram page 0
	return
;-----------------------
LongDelay           ;Approx 1 second delay
	movlw   d'250'
VarDelay
	movwf   DelCount2
LongDelLoop
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	decfsz  DelCount2
	goto    LongDelLoop
	return
;-------------------------
Delay1ms        ;5N+5 =  1250 clocks, * 2 = 2500 clocks = 1ms
	movlw   d'249'
	movwf   DelCount
Del1msLoop
	nop
	nop
	decfsz  DelCount
	goto    Del1msLoop
	movlw   d'249'
	movwf   DelCount
Del1ms2Loop
	nop
	nop
	decfsz  DelCount
	goto    Del1ms2Loop
	nop
	nop
	nop
	nop
	return
;------------------------
SetGrid			;Enter with address of EE data block with reg 0,1,2 details
	movwf	AutoEEADR	
	movlw	GRIDNUMBER		;Number of registers that have to be set each time, stored in block
	movwf	Counter
SetGridLoop
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx3

	incf	AutoEEADR
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx2

	incf	AutoEEADR
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx1

	incf	AutoEEADR
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx0

	call	Delay1ms
	call	SendSynth

	incf	AutoEEADR

	decfsz	Counter
	goto	SetGridLoop

	return
;------------------------
SendFromEE		;Send all registers that are not part of the frequency setting
			;first 4 registers defined by constants to save EE

;   de  0x00, 0x00, 0x00, 0x17    ;  Reg7   
;   de  0x00, 0x00, 0x00, 0xCD    ;  Reg13   
;   de  0x00, 0x00, 0x00, 0x1C    ;  Reg12   
;   de  0x28, 0x00, 0x14, 0x09    ;  Reg9   

	clrf	Tx3
	clrf	Tx2
	clrf	Tx1
	movlw	0x17
	movwf	Tx0
	call	Delay1ms
	call	SendSynth

	clrf	Tx3
	movlw	0x00		;Contains VCO_DIV_OPT, could be  0x04 See data sheet
	movwf	Tx2
	clrf	Tx1
	movlw	0xCD
	movwf	Tx0
	call	Delay1ms
	call	SendSynth

	clrf	Tx3
	clrf	Tx2
	clrf	Tx1
	movlw	0x17
	movwf	Tx0
	call	Delay1ms
	call	SendSynth

	movlw	0x28
	movwf	Tx3
	movlw	0x00
	movwf	Tx2
	movlw	0x14
	movwf	Tx1
	movlw	0x09
	movwf	Tx0
	call	Delay1ms
	call	SendSynth
			
			;Do the rest from EE
	movlw	LOW(ConfigData)
	movwf	AutoEEADR	
FromEELoop
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx3

	incf	AutoEEADR
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx2

	incf	AutoEEADR
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx1

	incf	AutoEEADR
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx0

	call	Delay1ms
	call	SendSynth

	incf	AutoEEADR
	movlw	LOW(EndConfigData)
	subwf	AutoEEADR, W		;W = Address - EndOfDataLabel, when 0/+Ve, C=1, exit sending
	btfss	STATUS, C	
	goto	FromEELoop

	return		
;-------------------------------
SendFreq		;Takes JTxx code 0 - nn, constructs register R0 and sends to synth chip
	movwf	Temp

	subwf	LastFreq, W	;Don't reprog the synth if same frequency is requested, 
	btfsc	STATUS, Z	;Avoids glitches due to chip responding to reprogging
	return

	movf	Temp, W
	movwf	LastFreq	;Update the current value

	movlw	LOW(JTFreqData + 3 + (GRIDNUMBER-1) * 4)		;N data in Reg0/1 remains unchanged.  F data in Reg2/3 is modified
	call	GetEE
	movwf	Tx0

	movlw	LOW(JTFreqData + 2 + (GRIDNUMBER-1) * 4)
	call	GetEE
	movwf	Tx1

	movlw	LOW(JTFreqData + 1 + (GRIDNUMBER-1) * 4)		
	call	GetEE
	movwf	Tx2

	movlw	LOW(JTFreqData + 0 + (GRIDNUMBER-1) * 4)
	call	GetEE
	movwf	Tx3

	movf	Temp, W		;*********   Only works if lowest 16 bits don't carry over.  NEED TO CHECK
	addwf	Tx2
	btfsc	STATUS, C
	incf	Tx3


SendSynth       ; ~~~~~~ Jump straight in here when other registers are sent from EE ~~~~~~~~

	bcf	SynthLE	
	movlw	SERIALBITS
	movwf	BitCount
SendChipLoop
	bcf	SynthClk
	rlf	Tx0
	rlf	Tx1
	rlf	Tx2
	rlf	Tx3		;Working bit in C
	btfsc	STATUS, C	
	bsf	SynthData		
	btfss	STATUS, C
	bcf	SynthData

	goto	$+1
	goto	$+1
	bsf	SynthClk	;Clocked in on rising edge

	decfsz	BitCount
	goto	SendChipLoop

	bcf	SynthClk	;Place into rest state
	bcf	SynthData
	nop
	bsf	SynthLE		;Needs to get back low in 1us
	bcf	SynthLE

	return


;-------------------------
GetGPSData          ;Monitors GPSIN line for Time Info
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  if GPSDataType ==0	;Oncore / M12 etc.  9600 baud,  starts @@Ea

	incf    SerCounter
	nop
	bsf     LEDGrn      ;Flash Green LED Until header is received.  If LEDRed
	btfss   GPSOk       ; is enabled, this gives a unique flash pattern
	bcf     LEDGrn      ;Flash in antiphase if GPS is not locked up
	call    PushStack
	call    GetGPSByte
	movwf   Stack0      ;String starts    @@Eamdyyhms
                            ;Stack position - A9876543210
	movf    StackA , W      
	sublw   "@"        
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack9 , W     
	sublw   "@"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack8 , W      
	sublw   "E"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack7 , W      
	sublw   "a"
	btfss   STATUS , Z
	goto    GetGPSData
				;Having reached here means correct data message
				; has started and HMS data is stored in stack
	movf    Stack1, W  
	movwf   Minutes
	movf    Stack0, W
	movwf   Seconds

	bcf     LEDGrn      ;Green off means time data has been decoded
	btfss   GPSOk
	bsf     LEDGrn      ;Flash in antiphase if GPS is not locked up

GetGPSStatus        ;Now read in rest of stack data until [cr][lf]
	incf    SerCounter      ;  at the end.  BUT, these may occur as part of
	call    PushStack       ;  binary data, so only test after 74 chars
	call    GetGPSByte     ;  have been received (should be 76, but...)
	movwf   Stack0     

	movlw   d'74'
	subwf   SerCounter, W
	btfss   STATUS, C       ;W = SerCounter - W, if +Ve (C=1) count is ready
	goto    GetGPSStatus    ;  to test for [cr][lf]

	movf    Stack0, W
	sublw   0x0A    ;Stack ends with    ...msdsC[cr][lf]
	btfss   STATUS , Z      ;              65432 1   0 
	goto    GetGPSStatus    ;At some future point, could validate 
	movf    Stack1, W       ; with a checksum calculation.
	sublw   0x0D
	btfss   STATUS , Z
	goto    GetGPSStatus
							;Reached here means terminating [cr][lf] is in
	movf    Stack3, W       ;  the right place, so extract status info
	movwf   RxStatus

	bcf     GPSOk
	movlw   b'00001010' ;Mask the GPS status data to give just the
	andwf   RxStatus, W ; bits that show no proper fix then, only
	btfsc   STATUS, Z   ; if these are all zero,  set GPSOk Flag
	bsf     GPSOk       ; WAS b'01001011', now ignores poor geometry flag
                        ;     and bad almanac 
	clrf    SerCounter      ;Ready for next time
	return
  endif
;--------------------------------------
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  if GPSDataType == 1		;NMEA Format data at 4800 baud

	nop
	bsf     LEDGrn      ;Flash Green LED Until header is received.  If LEDRed
	btfss   GPSOk       ; is enabled, this gives a unique flash pattern
	bcf     LEDGrn      ;Flash in antiphase if GPS is not locked up
	call    PushStack
	call    GetGPSByte
	movwf   Stack0		;String starts    $GPRMC,HHMMSS.ss   (decimal seconds optional)
				;Stack position - CBA9876543210
	movf    StackC , W      
	sublw   "$"        
	btfss   STATUS , Z
	goto    GetGPSData
	movf    StackB , W     
	sublw   "G"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    StackA , W      
	sublw   "P"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack9 , W      
	sublw   "R"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack8 , W      
	sublw   "M"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack7 , W      
	sublw   "C"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack6 , W      ;May be overkill checking all these
	sublw   ","		;   but... whatever....
	btfss   STATUS , Z
	goto    GetGPSData

	movf	Stack2, W	;String starts    $GPRMC,HHMMSS.ss   (decimal seconds optional)
	movwf	BCDLo		;Stack position - CBA9876543210
	movf	Stack3, W
	movwf	BCDHi
	call	BCDtoBIN
	movwf	Minutes

	movf	Stack0, W
	movwf	BCDLo
	movf	Stack1, W
	movwf	BCDHi
	call	BCDtoBIN
	movwf	Seconds

	bcf     LEDGrn      	;Green off means time data has been decoded
	btfss   GPSOk		;Flash in antiphase if GPS is not locked up
	bsf     LEDGrn      	

				;Now continue looking for the comma delimiting the status value
NextComma
	nop
	call    PushStack
	call    GetGPSByte	;After arbitrary seconds formatting...
	movwf	Stack0		;                 xxxx,A,    (or V) 
				;Stack position - 6543210
	sublw	","
	btfss	STATUS, Z
	goto	NextComma
	movf	Stack2, W
	sublw	","
	btfss	STATUS, Z
	goto	NextComma	;
	
	movf	Stack1, W
	sublw	"A"
	bcf	GPSOk
	btfsc	STATUS, Z
	bsf	GPSOk

	return
;----------------------------
BCDtoBIN                ;Converts pairs of ASCII BCD characters to
	movf    BCDHi , W   ;binary in W
	andlw   b'00001111' ;mask from ASCII to digit  = [BCDHi]
	movwf   Temp
	bcf     STATUS , C
	rlf     Temp
	rlf     Temp        ;Temp = 4 * [BCDHi]
	movlw   b'00001111'
	andwf   BCDHi , W   ;W = [BCDHi]
	addwf   Temp        ;Temp = 5 * [BCDHi]
	bcf     STATUS , C
	rlf     Temp
	movf    BCDLo , W
	andlw   b'00001111'
	addwf   Temp        ;Temp = 10 * BCDHi + BCDLo
	movf    Temp , W
	return
;-------------------------------
  endif

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;--------------------------------------
GetGPSByte
 if NMEAPol == 1        ; 1  =  RS232 levels (0/-ve = Idle
    btfss   GPSIN       ; 0  =  Logic levels (1 = Idle)
 else
    btfsc   GPSIN
 endif
    goto    GetGPSByte     ;Loop waiting for start bit

    call    NMEAHalfBitDel    ;Centre of start bit
 if NMEAPol == 1
    btfss   GPSIN
 else
    btfsc   GPSIN       ;Check it's still low
 endif
    goto    GetGPSByte

	movlw   8    
	movwf   Counter
    clrf    Temp        ;probably not needed, but good practice

InitByteLoop
    call    NMEAHalfBitDel
    call    NMEAHalfBitDel    ;
	nop
	bcf     STATUS , C      ;start in known state

 if NMEAPol == 1
    btfss   GPSIN
 else
    btfsc   GPSIN
 endif
	bsf     STATUS , C      ;Data into Carry
    rrf     Temp        ;and then into Temp, LSB first
	decfsz  Counter
    goto    InitByteLoop    ;loop is 8 + 2 * NMEAHalfBitDel long
	 
    call    NMEAHalfBitDel    ;#### Don't really need to wait for stop bit! ####
							; Now have approx 50us of stop bit left
    movf    Temp , W    ;to use this data

    return          ;Exit with received byte in W and Temp
;-----------------------------------------


;-----------------------------------------
NMEAHalfBitDel     ;Total delay 3.N + 5 clocks

  if GPSDataType == 0		;Oncore 9600 baud
	movlw   BAUD9600   
  else			
	movlw	BAUD4800	;NMEA 4800 baud 
  endif			;  

	movwf   DelCount
BaudDelLoop
	decfsz  DelCount
	goto    BaudDelLoop
	return
;-------------------------------
PushStack
	movf    StackE , W
	movwf   StackF
	movf    StackD , W  
	movwf   StackE
	movf    StackC , W
	movwf   StackD
	movf    StackB , W
	movwf   StackC
	movf    StackA , W
	movwf   StackB
	movf    Stack9 , W
	movwf   StackA
	movf    Stack8 , W
	movwf   Stack9
	movf    Stack7 , W
	movwf   Stack8
	movf    Stack6 , W
	movwf   Stack7 
	movf    Stack5 , W
	movwf   Stack6
	movf    Stack4 , W
	movwf   Stack5
	movf    Stack3 , W
	movwf   Stack4
	movf    Stack2 , W
	movwf   Stack3
	movf    Stack1 , W
	movwf   Stack2
	movf    Stack0 , W
	movwf   Stack1

	return
;-------------------------------
SendCWMsg           ;Enter with EE address programmed into CWMsgCnt
	movf    CWMsgCnt, W
    	call    GetEE       ;Test for terminator at start = no CW message
	addlw   0
	btfsc   STATUS, Z
	return

	bcf	KEYLINE 
	call    CWDelay
	call    CWDelay
	call    CWDelay
CWLoop
	movf    CWMsgCnt, W
	call    GetEE
	addlw   0
	btfsc   STATUS, Z
	goto    DoneCWMsg

	call    SendCWChar
	incf    CWMsgCnt
	goto    CWLoop
DoneCWMsg
	call    CWDelay
	call    CWDelay
	call    CWDelay
	bsf	KEYLINE 	;Exit with key down
	return
;--------------------------
SendCWChar      ;Sends a single CW character
	addlw   0xE0     ;Subtract 32 (add 256 - 32) to point to table
	call    CWTable
	movwf   CWChar      ;Doesn't affect STATUS so...
	movf    CWChar      ;  check for invalid returned code here
	btfsc   STATUS , Z      
	goto    NoChar

	movlw   0x07    
	andwf   CWChar , W      ;Extract element count into W
	movwf   CWCharCnt     
        movlw   6       
	subwf   CWCharCnt , W     ;W = Elcount - 6.  If +ve/0 then C=1, > 5 els
	btfss   STATUS , C
	goto    Send5
	
	rrf     CWCharCnt   ;Puts LSB into C as 110 = .  111 = -
	call    SendEle     ;First  . or -  for 6 element characters
	movlw   5
	movwf   CWCharCnt     ;Force count to 5 for rest of the elements
Send5       
	rlf     CWChar      ;Element to be sent into C.
	call    SendEle     ;   count already in Counter
	decfsz  CWCharCnt   
	goto    Send5
NoChar          ;Inter word gap, and invalid codes.
	bcf     LEDRed
	bcf	KEYLINE 
	call    CWDelay
	call    CWDelay
	return
;-------------------
SendEle         ;Data in STATUS , C;  0 = dit  1 = dah
	rrf     Stack0      ;Save C bit temporarily on stack

	bsf     LEDRed
	bsf	KEYLINE 
	rlf     Stack0      ;Recovery C bit again, to get dot/dash 
	btfss   STATUS , C      ;test for dot or dash
	goto    SendDot
	call    CWDelay
	call    CWDelay
SendDot
	call    CWDelay
	bcf     LEDRed
	bcf	KEYLINE 
	call    CWDelay     ;Inter element gap
	return
;-------------------
CWDelay       
	movlw   CWSPEED
	movwf   DelCount2
CWDelLoop
	call    Delay1ms      
	decfsz  DelCount2
	goto    CWDelLoop
	return
;--------------------------
UpCase
    return

;-------------------------
DecodeHex   ;Single ASCII character to binary with error checking
        ;Untrapped characters remaining ...........
        ;at the moment Upper Case letters G-Z plus characters
        ;  : ; < = > ? @ [ \ ] ^ _ `    will be let through.
        ;Chars 0 - $2F and letters "g" up
        ;are detected.

        bsf     HexFlag         ;Flag for valid hex character
        movwf   Temp            ;
        movlw   0x30
        subwf   Temp            ;subtract $30 and leave in Temp
        btfss   STATUS , C      ;if < $30, result -ve so C=0, error, set=0
        bcf     HexFlag
        
        movf    Temp , W        ;Test for residual from above > d54 ie > "g"
        sublw   d'54'           ;W = 54 - W, if < 54  invalid char, -ve, C=0
        btfss   STATUS , C      ;
        bcf     HexFlag

        movf    Temp , W
        andlw   0x1F            ;mask to 5 bits, resolves upper / lower case
        movwf   Temp            ;if 0 - 9 then Temp is correct, else 7 high

        movf    Temp , W        
        sublw   d'9'            ;W = 9 - W
        btfsc   STATUS , C      ;if -ve, C=0, char > 9, so subtract 7
        goto    ConvDone
        movlw   d'7'
        subwf   Temp            ;
ConvDone
        movf    Temp , W        ;
        andlw   0x0F            ;mask out any rubbish left in
        return                  ;flag set if partly valid 
;-------------------------

;-----------------------------------------
SendGPSByte                    
	movwf   Temp
 if NMEAPol == 1
    bsf     GPSOUT
 else
    bcf     GPSOUT             ;start bit
 endif
    call    NMEAHalfBitDel        ;
    call    NMEAHalfBitDel
	movlw   8
	movwf   Counter
SendGPSDatLoop
	rrf     Temp                ;bit into C


 if NMEAPol == 1
	btfsc   STATUS , C
    bcf     GPSOUT
	btfss   STATUS , C
    bsf     GPSOUT             ;could have 2us jitter depending on 1/0
 else
	btfsc   STATUS , C
    bsf     GPSOUT
	btfss   STATUS , C
    bcf     GPSOUT             ;could have 2us jitter depending on 1/0
 endif

    call    NMEAHalfBitDel
    call    NMEAHalfBitDel
	decfsz  Counter
    goto    SendGPSDatLoop         ;loop 8 + BitDelay long

 if NMEAPol == 1
    bcf     GPSOUT
 else
    bsf     GPSOUT
 endif
    call    NMEAHalfBitDel    
    call    NMEAHalfBitDel        ;Two stop bits
    call    NMEAHalfBitDel
    call    NMEAHalfBitDel

	return
;-----------------------------------------
SetUpGPS
	bsf	GPSOUT		;Send on off to A2200 module
	call	Delay1ms
	bcf	GPSOUT
	call	Delay1ms
	bsf	GPSOUT		;
	call	Delay1ms
	bcf	GPSOUT
	return

;---------------------------------------
;============================================
org 0x300

CWTable       ;CW Characters in compressed form. 
	movwf   Temp
	movlw   HIGH($)
    movwf   PCLATH    
	movf    Temp , W
	addwf   PCL
	
	retlw   b'00000000'  ; [sp] 00
	retlw   b'00000000'  ;  !   00
	retlw   b'10010110'  ;  "   96
	retlw   b'00010101'  ;  #   15
	retlw   b'00000000'  ;  $   00
	retlw   b'00000000'  ;  %   00
	retlw   b'00000000'  ;  &   00
	retlw   b'11110110'  ;  '   F6
	retlw   b'10110101'  ;  (   B5
	retlw   b'01101111'  ;  )   6F
	retlw   b'01000101'  ;  *   45
	retlw   b'00000000'  ;  +   00
	retlw   b'10011111'  ;  ,   9F     Coding for CW characters of up to
	retlw   b'00001111'  ;  -   0F      6 elements.  Stored as :
	retlw   b'10101110'  ;  .   AE     Bits 7 to 3 make first five CW elements
	retlw   b'10010101'  ;  /   95     Reading  Left to Right 
	retlw   b'11111101'  ;  0   FD     Coded as 0 = dit   1 = dah
	retlw   b'01111101'  ;  1   7D     Bts 2-0 are number of elements except for 
	retlw   b'00111101'  ;  2   3D     110 > 1st element is a dit  
	retlw   b'00011101'  ;  3   1D     111 > 1st element is a dah       
	retlw   b'00001101'  ;  4   0D
	retlw   b'00000101'  ;  5   05
	retlw   b'10000101'  ;  6   85
	retlw   b'11000101'  ;  7   C5
	retlw   b'11100101'  ;  8   E5
	retlw   b'11110101'  ;  9   F5
	retlw   b'00000000'  ;  :   00
	retlw   b'00000000'  ;  ;   00
	retlw   b'00000000'  ;  <   00
	retlw   b'10001101'  ;  =   8D
	retlw   b'00000000'  ;  >   00
	retlw   b'01100110'  ;  ?   66
	retlw   b'11010110'  ;  @   D6
	retlw   b'01000010'  ;  A   42
	retlw   b'10000100'  ;  B   84
	retlw   b'10100100'  ;  C   A4
	retlw   b'10000011'  ;  D   83
	retlw   b'00000001'  ;  E   01
	retlw   b'00100100'  ;  F   24
	retlw   b'11000011'  ;  G   C3
	retlw   b'00000100'  ;  H   04
	retlw   b'00000010'  ;  I   02
	retlw   b'01110100'  ;  J   74
	retlw   b'10100011'  ;  K   A3
	retlw   b'01000100'  ;  L   44
	retlw   b'11000010'  ;  M   C2
	retlw   b'10000010'  ;  N   82
	retlw   b'11100011'  ;  O   E3
	retlw   b'01100100'  ;  P   64
	retlw   b'11010100'  ;  Q   D4
	retlw   b'01000011'  ;  R   43
	retlw   b'00000011'  ;  S   03
	retlw   b'10000001'  ;  T   81
	retlw   b'00100011'  ;  U   23
	retlw   b'00010100'  ;  V   14
	retlw   b'01100011'  ;  W   63
	retlw   b'10010100'  ;  X   94
	retlw   b'10110100'  ;  Y   B4
	retlw   b'11000100'  ;  Z   C4


;================================

    cblock 0x20
        Temp
	AutoEEADR
        DelCount, DelCount2
        IntCountLo,  IntCountHi
        Counter
        SymbCount       ;Counts 0 to 126 for each JT65 symbol
        CharCount       ;
	JT4Data
	LastFreq
	PosData
        Flags
        Wreg
        SaveSts
        Seconds, Minutes
	BCDHi, BCDLo
	A0, A1, A2, A3
	B0, B1, B2, B3
	D0, D1, D2, D3
	Q0, Q1, Q2, Q3, Q4, Q5, Q6, Q7
	FreqCache
        ICount
	BitCount
	Tx0, Tx1, Tx2, Tx3
        CommaCount
        CWChar, CWMsgCnt, CWCharCnt
        RxChar
        TonePtr
        SerCounter      ;Count of characters on GPS serial data
        RxStatus    ;GPS Receiver status
        AddrHold
        MsgCount
        Stack0, Stack1, Stack2, Stack3, Stack4, Stack5, Stack6, Stack7 
	Stack8, Stack9, StackA, StackB, StackC, StackD, StackE, StackF
    endc   
;==========================================
		end
