;JT65GEN_FREERUN_LMX2470.ASM    G4JNT  Sept 2012
;JT65 Symbol generator for driving FractN Synth directly
; Derived from JT4FreeRun
;
;Uses flexible interrupt timing by reprog of TMR0 register 

    LIST P = 16F627A

        INCLUDE "P16F627A.INC"
    __config    0x3F62      ;LV Programming off , OSC HS
    __idlocs    0x8040      ;


	#define		SynthData	PORTA, 0
	#define		SynthClk	PORTA, 1
	#define		SynthLE		PORTA, 2
	#define     	KEYLINE     	PORTA, 3
    
 
	#define     	TESTLINK    	PORTB, 3
	#define     	LEDGrn      	PORTB, 4
	#define     	LEDRed      	PORTB, 5

	#define     	TXENABLE    	PORTB, 6  
	#define     	CLOCKTEST  	PORTB, 7


	#define     	TestModeFlag    Flags, 1
	#define	    	AllowData	Flags, 2
	#define     	AllowSendFlag	Flags, 3
	#define     	CurrData        Flags, 4
	#define     	HexFlag         Flags, 6


	SERIALBITS	=	d'24'
	BOTHMINUTES	=	1	;1 = every minute, 0 = Alternate minutes
    	CWSPEED     	=   	d'50'           ;CW Dot length, ms


			;Values for 10MHz xtal

	INTDIVIDER	=	d'134'		;value X, to generate Interupt divisor N (see below)
	JTDIVIDER	=	d'1858'		;  = INTSPERSECOND / Symbol rate
	INTSPERSECOND	=	d'5000'		;  = Fxtal / 2000  		
	MSDELAY		=	d'124'		;  = Fxtal / 80000 - 1 

  errorlevel 1, -207  ;Suppress msgs. and warnings about labels after col 1.
						   

	org     0
    nop
    movlw   b'00000000'     ;No interupts until passed thro here once
    movwf   INTCON          ;
    goto    StartUp         ;jump to main code

;------------------ Interrupt routine, get here 4 cycles after TMR overflow --
ints                           
	bcf     INTCON , T0IF   ;		5kHz interrupt
        movwf   Wreg        ;Save W
	swapf   STATUS , W      ; nb.  movwf does not affect any STATUS bits
	movwf   SaveSts         ;Stave STATUS, without affecting it

	bsf	CLOCKTEST
	movlw	INTDIVIDER
	nop			;nops needed to pad out so total clocks here 
				;  are a multiple of the prescalar value 

	movwf	TMR0		;TMR0 start recounting from here

		;Interrupt counter goes from X to 256 at Fclock / Prescale.  
		;   Let the overhead = V clocks, then int occurs every (256 - X) * P + V   clocks
		;  X =  256 - (N-V)/P    V must be padded out to be integer multiple of P

;   Below here is not time - critical any more. But must complete in remaining interrupt interval
	btfss	AllowData
	goto	KeepTime
	
	incf    IntCountLo
	btfsc   STATUS, Z
	incf    IntCountHi

        movlw   HIGH(JTDIVIDER)    ;5000/JTDIVIDER = symbol rate
	subwf   IntCountHi, W
	btfss   STATUS, Z
	goto    KeepTime

        movlw   LOW(JTDIVIDER)
	subwf   IntCountLo, W
	btfss   STATUS, Z
	goto    KeepTime

	   ;............             Here at 2.69Hz
	clrf    IntCountLo
	clrf    IntCountHi

	btfsc   SymbCount, 0    ;Flash Red LED at half symbol rate
	bsf     LEDRed
	btfss   SymbCount, 0
	bcf     LEDRed

	movf    SymbCount       ;Test for First symbol, in which case
	btfsc   STATUS, Z       ;  don't do any update
	goto    NotFirstPass

NotFirstPass

	movlw   d'126'
	subwf   SymbCount, W    ;So do the testing for a full message here, 
	btfss   STATUS, Z
	goto    NotLastSymbol

	bcf	AllowData	
	goto    KeepTime    ;  and don't do anything else
	;...............
NotLastSymbol
		movf    SymbCount       ;Test for first symbol
		btfsc   STATUS, Z
		call    SyncToStack     ;Load sync pattern into stack first time here

		rlf     Stack0          ;Stack holds sync / message tone pattern
		rlf     Stack1    
		rlf     Stack2
		rlf     Stack3          
		rlf     Stack4
		rlf     Stack5
		rlf     Stack6
		rlf     Stack7
		rlf     Stack8
		rlf     Stack9
		rlf     StackA
		rlf     StackB
		rlf     StackC
		rlf     StackD
		rlf     StackE
		rlf     StackF
		btfss   STATUS, C       ;1 = sync, 0 = character
		goto    JTSymb

		clrf	FreqCache            ;0 passed to routine sends sync tone
		bsf	AllowSendFlag
		goto    SendTone
	;...................
JTSymb                          ;Get next symbol value in message (0-63)
	movf    CharCount, W
	addlw   LOW(JT65MsgData)
	call    GetEE           ;Get value 0 to 63
        addlw   2               ;2* interval Offset from sync (was 3 in JT44)
	movwf	FreqCache
	bsf	AllowSendFlag   ;Pass symbol value to routine,
								; generates frequency data in D0-D3
	incf    CharCount
	;..........
SendTone                        ;Get here with frequency data in D0-5
	incf    SymbCount       ;Increment here, but tested at start allows 
								; the DDS update at last+1 interval, ie. =136
KeepTime
	incf	TimCountLo
	btfsc	STATUS, Z
	incf	TimCountHi

	movlw	HIGH(INTSPERSECOND)
	subwf	TimCountHi, W
	btfss	STATUS, Z
	goto	OutInterrupt

	movlw	LOW(INTSPERSECOND)
	subwf	TimCountLo, W
	btfss	STATUS, Z
	goto	OutInterrupt

	clrf	TimCountHi
	clrf	TimCountLo

	incf	Seconds
	movlw	d'60'
	subwf	Seconds, W
	btfss	STATUS, Z
	goto	OutInterrupt
	clrf	Seconds
	incf	Minutes		;Minutes just roll 0 - 255

OutInterrupt
	bcf	CLOCKTEST                              
	swapf   SaveSts , W     ;Restore STATUS
	movwf   STATUS
	swapf   Wreg            ;Restore W
	swapf   Wreg , W        ;Can't use movf as it  affects Z bit
			
	retfie		;Vaguely 90 clocks or so in here, so OK for 125 division
;=====================================
StartUp 
    clrf    PORTA       ;Special setup to disable comparator on 'F628
    movlw   7
    movwf   CMCON
	
    bsf     STATUS,RP0      ;ram page 1
    movlw   b'00000001'     ;Internal input, Prescalar /4  pull-ups
    movwf   OPTION_REG
    movlw   b'00000000'     ;
    movwf   PORTA           ;
    movlw   b'00001011'     ; B3 TESTLINK 
    movwf   PORTB           ;B7 Clock Test output
    bcf     STATUS,RP0      ;ram page 0
 
    call    SendFromEE	    ;Initialise other registers

    clrf    PORTB
    clrf    PORTA
    clrf    TMR0
    clrf    IntCountHi
    clrf    IntCountLo
    clrf    SymbCount
    clrf    Flags
    clrf    Seconds	  
    clrf    Minutes
    clrf    CharCount

			;Launch straight in to sending, assuming unit has been reset at the minute point
    bcf     LEDGrn
    movlw   b'10000000'     ;Enable global int, not timer overflow int yet
    movwf   INTCON
    bsf     INTCON, T0IE
;...........................
MainLoop
    btfss   TESTLINK
    call    TestMode

  if BOTHMINUTES == 0

    btfsc   Minutes, 0      ;Test for the odd minute
    goto    CarryOn

  endif

    movf    Seconds, W		;At start, this happens even before the interrupt has updated the time
    sublw   d'00'
    btfss   STATUS, Z
    goto    CarryOn 
                      ;Now at even minute : 00

    bsf     TXENABLE    ;Enable Tx up to 1s earlier to allow for relays
    call    LongDelay	;Start at about 1s interval
    bsf     KEYLINE
    bcf     LEDGrn
    movlw   0        	;Set Tone 0
    call    SendFreq

    clrf    CharCount    
    clrf    IntCountLo
    clrf    IntCountHi
    clrf    SymbCount    ;Re-Initialise registers

    bsf     AllowData	 ;Starts JT4 sequence

JT65Loop
    btfss   AllowSendFlag
    goto    NoSend
    movf    FreqCache, W
    call    SendFreq		;Has to be done outside interrupt routine
    bcf	    AllowSendFlag	; as an interrupt overrun would occur
NoSend
    btfsc   AllowData	 ;Sit in here waiting for completion   
    goto    JT65Loop

    movlw   0        	;Set Tone 0
    call    SendFreq

    bcf     KEYLINE
    call    LongDelay

    movlw   LOW(CWMsg)
    movwf   CWMsgCnt
    call    SendCWMsg

    bcf     TXENABLE
    bsf     LEDGrn
    bcf	    LEDRed
CarryOn
    goto    MainLoop
;.............................

;===========================================
TestMode                    ;TESTLINK Input grounded
    bsf     LEDRed
    bcf     LEDGrn
    movlw   0        	;Set Tone 0
    call    SendFreq
TestModeLoop
    btfss   TESTLINK
    goto    TestModeLoop

    bcf     LEDRed
    bcf     TXENABLE
    bcf     KEYLINE
    return
;--------------------------------

GetEE         ;different from ;F84   EE registers all in bank 1 now 
	bsf     STATUS,RP0      ;ram page 1
	movwf   EEADR
	bsf     EECON1 , RD
	movf    EEDATA , W
	bcf     STATUS , RP0    ;ram page 0
	return
;-----------------------
LongDelay           ;Approx 1 second delay
	movlw   d'250'
	movwf   DelCount2
LongDelLoop
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	decfsz  DelCount2
	goto    LongDelLoop
	return
;-------------------------
Delay1ms        ;20N+5
	movlw   MSDELAY
	movwf   DelCount
Del1msLoop
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	nop
	decfsz  DelCount
	goto    Del1msLoop
	return
;-------------------------


;-------------------------------
PushStack
	movf    StackE , W
	movwf   StackF
	movf    StackD , W  
	movwf   StackE
	movf    StackC , W
	movwf   StackD
	movf    StackB , W
	movwf   StackC
	movf    StackA , W
	movwf   StackB
	movf    Stack9 , W
	movwf   StackA
	movf    Stack8 , W
	movwf   Stack9
	movf    Stack7 , W
	movwf   Stack8
	movf    Stack6 , W
	movwf   Stack7 
	movf    Stack5 , W
	movwf   Stack6
	movf    Stack4 , W
	movwf   Stack5
	movf    Stack3 , W
	movwf   Stack4
	movf    Stack2 , W
	movwf   Stack3
	movf    Stack1 , W
	movwf   Stack2
	movf    Stack0 , W
	movwf   Stack1

	return
;-------------------------------
SendCWMsg           ;Enter with EE address programmed into CWMsgCnt
	movf    CWMsgCnt, W
    call    GetEE       ;Test for terminator at start = no CW message
	addlw   0
	btfsc   STATUS, Z
	return
    bcf     KEYLINE
	call    CWDelay
	call    CWDelay
	call    CWDelay
	call    CWDelay

CWLoop
	movf    CWMsgCnt, W
	call    GetEE
	addlw   0
	btfsc   STATUS, Z
	goto    DoneCWMsg

	call    SendCWChar
	incf    CWMsgCnt
	goto    CWLoop
DoneCWMsg
	call    CWDelay
	call    CWDelay
	call    CWDelay
	call    CWDelay
	return
;--------------------------
SendCWChar      ;Sends a single CW character
	addlw   0xE0     ;Subtract 32 (add 256 - 32) to point to table
	call    CWTable
	movwf   CWChar      ;Doesn't affect STATUS so...
	movf    CWChar      ;  check for invalid returned code here
	btfsc   STATUS , Z      
	goto    NoChar

	movlw   0x07    
	andwf   CWChar , W      ;Extract element count into W
	movwf   CWCharCnt     
    movlw   6       
	subwf   CWCharCnt , W     ;W = Elcount - 6.  If +ve/0 then C=1, > 5 els
	btfss   STATUS , C
	goto    Send5
	
	rrf     CWCharCnt   ;Puts LSB into C as 110 = .  111 = -
	call    SendEle     ;First  . or -  for 6 element characters
	movlw   5
	movwf   CWCharCnt     ;Force count to 5 for rest of the elements
Send5       
	rlf     CWChar      ;Element to be sent into C.
	call    SendEle     ;   count already in Counter
	decfsz  CWCharCnt   
	goto    Send5
NoChar          ;Inter word gap, and invalid codes.
	bcf     LEDRed
	call    CWDelay
	call    CWDelay
	return
;-------------------
SendEle         ;Data in STATUS , C;  0 = dit  1 = dah
	rrf     Stack0      ;Save C bit temporarily on stack
    bsf     KEYLINE
	bsf     LEDRed
	rlf     Stack0      ;Recovery C bit again, to get dot/dash 
	btfss   STATUS , C      ;test for dot or dash
	goto    SendDot
	call    CWDelay
	call    CWDelay
SendDot
	call    CWDelay
    bcf     KEYLINE
	bcf     LEDRed
    call    CWDelay     ;Inter element gap
	return
;-------------------
CWDelay       
	movlw   CWSPEED
	movwf   DelCount2
CWDelLoop
	call    Delay1ms      
	decfsz  DelCount2
	goto    CWDelLoop
	return
;-------------------------
;-------------------------
SendFreq		;Takes JTxx code 0 - nn, constructs register R0 and sends to synth chip
	movwf	Temp		
	bcf	STATUS, C
	rlf	Temp		;Align for Reg0 position,  with LSB = address bit.  Use more
				; RLFs when FractN resolution is a submultiple of tone spacing
			
	movlw	LOW(Reg0Base + 2)	;LSB of F data	, All shifted 1 bit left for Reg0 address
	call	GetEE		;(GetEE doesn't corrupt Temp)
	movwf	Tx0
	movlw	LOW(Reg0Base + 1)	
	call	GetEE		
	movwf	Tx1		;Tx0/1 contains 12 LSBs of F data and portion of N, 1 bit shifted left

	movf	Temp, W		;Add in (1 bit left shifted) offset for JTxx data
	addwf	Tx0
	btfsc	STATUS, C
	incf	Tx1		;This will fail if the 12 bits overflow - check when calculating values

	movlw	LOW(Reg0Base + 0)	
	call	GetEE		
	movwf	Tx2		;Remaining N data

SendSynth       ; ~~~~~~ Jump straight in here when other registers are sent from EE ~~~~~~~~

	bcf	SynthLE	
	movlw	SERIALBITS
	movwf	BitCount
SendChipLoop
	bcf	SynthClk
	rlf	Tx0
	rlf	Tx1
	rlf	Tx2		;Working bit in C
	btfsc	STATUS, C	
	bsf	SynthData		
	btfss	STATUS, C
	bcf	SynthData

	goto	$+1
	goto	$+1
	bsf	SynthClk	;Clocked in on rising edge

	decfsz	BitCount
	goto	SendChipLoop

	bcf	SynthClk	;Place into rest state
	bcf	SynthData
	nop
	bsf	SynthLE		;Needs to get back low in 1us
	bcf	SynthLE

	return
;------------------------------
SendFromEE		;Send all registers that are not part of the frequency setting
	movlw	LOW(ConfigData)
	movwf	AutoEEADR	
FromEELoop
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx2
	incf	AutoEEADR
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx1
	incf	AutoEEADR
	movf	AutoEEADR, W
	call	GetEE
	movwf	Tx0
	call	Delay1ms
	call	SendSynth

	incf	AutoEEADR
	movlw	LOW(EndConfigData)
	subwf	AutoEEADR, W		;W = Address - EndOfDataLabel, when 0/+Ve, C=1, exit sending
	btfss	STATUS, C	
	goto	FromEELoop

	return
;----------------------------------------------------------------------------------------
;---------------------------
SyncToStack
	movlw   0x98
	movwf   StackF
	movlw   0xFD
	movwf   StackE
	movlw   0x45
	movwf   StackD
	movlw   0x91
	movwf   StackC
	movlw   0xCF
	movwf   StackB
	movlw   0x6F
	movwf   StackA
	movlw   0x1A
	movwf   Stack9
	movlw   0xB3
	movwf   Stack8
	movlw   0x54
	movwf   Stack7
	movlw   0x81
	movwf   Stack6
	movlw   0x80
	movwf   Stack5
	movlw   0xD2
	movwf   Stack4
	movlw   0xD5
	movwf   Stack3
	movlw   0x32
	movwf   Stack2
	movlw   0x43
	movwf   Stack1
	movlw   0xFC
	movwf   Stack0

	return
;---------------------------------------
;============================================


org 0x300

CWTable       ;CW Characters in compressed form. 
	movwf   Temp
	movlw   HIGH($)
    movwf   PCLATH    
	movf    Temp , W
	addwf   PCL
	
	retlw   b'00000000'  ; [sp] 00
	retlw   b'00000000'  ;  !   00
	retlw   b'10010110'  ;  "   96
	retlw   b'00010101'  ;  #   15
	retlw   b'00000000'  ;  $   00
	retlw   b'00000000'  ;  %   00
	retlw   b'00000000'  ;  &   00
	retlw   b'11110110'  ;  '   F6
	retlw   b'10110101'  ;  (   B5
	retlw   b'01101111'  ;  )   6F
	retlw   b'01000101'  ;  *   45
	retlw   b'00000000'  ;  +   00
	retlw   b'10011111'  ;  ,   9F     Coding for CW characters of up to
	retlw   b'00001111'  ;  -   0F      6 elements.  Stored as :
	retlw   b'10101110'  ;  .   AE     Bits 7 to 3 make first five CW elements
	retlw   b'10010101'  ;  /   95     Reading  Left to Right 
	retlw   b'11111101'  ;  0   FD     Coded as 0 = dit   1 = dah
	retlw   b'01111101'  ;  1   7D     Bts 2-0 are number of elements except for 
	retlw   b'00111101'  ;  2   3D     110 > 1st element is a dit  
	retlw   b'00011101'  ;  3   1D     111 > 1st element is a dah       
	retlw   b'00001101'  ;  4   0D
	retlw   b'00000101'  ;  5   05
	retlw   b'10000101'  ;  6   85
	retlw   b'11000101'  ;  7   C5
	retlw   b'11100101'  ;  8   E5
	retlw   b'11110101'  ;  9   F5
	retlw   b'00000000'  ;  :   00
	retlw   b'00000000'  ;  ;   00
	retlw   b'00000000'  ;  <   00
	retlw   b'10001101'  ;  =   8D
	retlw   b'00000000'  ;  >   00
	retlw   b'01100110'  ;  ?   66
	retlw   b'11010110'  ;  @   D6
	retlw   b'01000010'  ;  A   42
	retlw   b'10000100'  ;  B   84
	retlw   b'10100100'  ;  C   A4
	retlw   b'10000011'  ;  D   83
	retlw   b'00000001'  ;  E   01
	retlw   b'00100100'  ;  F   24
	retlw   b'11000011'  ;  G   C3
	retlw   b'00000100'  ;  H   04
	retlw   b'00000010'  ;  I   02
	retlw   b'01110100'  ;  J   74
	retlw   b'10100011'  ;  K   A3
	retlw   b'01000100'  ;  L   44
	retlw   b'11000010'  ;  M   C2
	retlw   b'10000010'  ;  N   82
	retlw   b'11100011'  ;  O   E3
	retlw   b'01100100'  ;  P   64
	retlw   b'11010100'  ;  Q   D4
	retlw   b'01000011'  ;  R   43
	retlw   b'00000011'  ;  S   03
	retlw   b'10000001'  ;  T   81
	retlw   b'00100011'  ;  U   23
	retlw   b'00010100'  ;  V   14
	retlw   b'01100011'  ;  W   63
	retlw   b'10010100'  ;  X   94
	retlw   b'10110100'  ;  Y   B4
	retlw   b'11000100'  ;  Z   C4


;================================
  org   0x2100
ConfigData		;24 bit words sent in sequence  for registers that remain unchanged
  ;  Fout  2320.349271MHz   Resolution 5.3833Hz
  ;  R =  1      D =  1857596      N =  232      F =  64880
;   de  0x1D, 0x1A, 0xE0    ;  Reg0   N + F (LSBs)
;   de  0x41, 0x83, 0xC1    ;  Reg1   R + D (LSBs)
;   de  0xC0, 0x07, 0x43    ;  Reg2   IF Disabled
;   de  0x38, 0x00, 0x15    ;  Reg3   Charge Pump Current
;   de  0xBC, 0x2B, 0xC7    ;  Reg4   Timeout / FastLock
;   de  0x10, 0x1F, 0x49    ;  Reg5   IF not used
;   de  0x00, 0xC4, 0xFB    ;  Reg6   Config and setup bits
;   de  0x71, 0x40, 0xFD    ;  Reg7   D(MSBs) + F (MSBs)
;   de  0x0C, 0x08, 0x6F    ;  Reg8   Dither settings


 ;  Fout 1296.34927MHz   Resolution 5.3833008Hz
  ;  R =  1      D =  1857596      N =  129      F =  1179438
   de  0x10, 0x3E, 0x5C    ;  Reg0   N + F (LSBs)
   de  0x41, 0x83, 0xC1    ;  Reg1   R + D (LSBs)
   de  0xC0, 0x07, 0x43    ;  Reg2   IF Disabled
   de  0x38, 0x00, 0x15    ;  Reg3   Charge Pump Current
   de  0xBC, 0x2B, 0xC7    ;  Reg4   Timeout / FastLock
   de  0x10, 0x1F, 0x49    ;  Reg5   IF not used
   de  0x00, 0xC4, 0xFB    ;  Reg6   Config and setup bits
   de  0x71, 0x51, 0xFD    ;  Reg7   D(MSBs) + F (MSBs)
   de  0x00, 0x08, 0x6F    ;  Reg8   Dither settings

EndConfigData

Reg0Base	
;   de  0x1D, 0x1A, 0xE0    ;  Reg0   N + F (LSBs);Reg0 needed for Sync frequency
   de  0x10, 0x3E, 0x5C    ;  Reg0   N + F (LSBs)

JT65MsgData
 include "jt65symb.inc"     ;Left Justified JT4 tones, 4 per byte

CWMsg           de      "G4JNT",0
;--------------------------------
    cblock 0x20
        Temp
	AutoEEADR
        DelCount, DelCount2
        IntCountLo,  IntCountHi
	TimCountLo, TimCountHi
        Counter
        SymbCount       ;Counts 0 to 126 for each JT4 symbol
	CharCount
        Amplitude
        Flags
        Wreg
        SaveSts
        Seconds, Minutes
	BCDHi, BCDLo
        ICount
	BitCount
	Tx0, Tx1, Tx2, Tx3
	F0, F1
	N0, N1
        CommaCount
        CWChar, CWMsgCnt, CWCharCnt
        RxChar
	FreqCache
        TonePtr
        SerCounter      
        AddrHold
        MsgCount
        Stack0, Stack1, Stack2, Stack3, Stack4, Stack5, Stack6, Stack7 
	Stack8, Stack9, StackA, StackB, StackC, StackD, StackE, StackF
    endc   
;==========================================
		end
